#+
#   VIDE -- Void IDentification and Examination -- ./python_tools/vide/apTools/chi2/cosmologyTools.py
#   Copyright (C) 2010-2014 Guilhem Lavaux
#   Copyright (C) 2011-2014 P. M. Sutter
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; version 2 of the License.
# 
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#+
# a suite of functions to compute expansion rates, angular diameter 
# distances, and expected void stretching

import numpy as np
import healpy as healpy
import os
from backend import *

__all__=['getSurveyProps', 'getNside', 'figureOutMask', 'findEdgeGalaxies']

# -----------------------------------------------------------------------------
# returns the volume and galaxy density for a given redshit slice
def getSurveyProps(maskFile, zmin, zmax, selFunMin, selFunMax, portion, 
                   omegaM, selectionFuncFile=None, useComoving=False):

  #LIGHT_SPEED = 299792.458

  mask = healpy.read_map(maskFile)
  area = (1.*np.size(np.where(mask > 0)) / np.size(mask)) * 4.*np.pi

  if useComoving:
    zmin = LIGHT_SPEED/100.*angularDiameter(zmin, Om=omegaM)
    zmax = LIGHT_SPEED/100.*angularDiameter(zmax, Om=omegaM)
  else:
    zmin = zmin * LIGHT_SPEED/100.
    zmax = zmax * LIGHT_SPEED/100. 

  if selectionFuncFile == None:
    nbar = 1.0

  elif not os.access(selectionFuncFile, os.F_OK):
    print("   Warning, selection function file %s not found, using default of uniform selection." % selectionFuncFile)
    nbar = 1.0

  else:
    selfunc = np.genfromtxt(selectionFuncFile)
    selfunc = np.array(selfunc)
    selfunc[:,0] = selfunc[:,0]/100.
    selfuncUnity = selfunc
    selfuncUnity[:,1] = 1.0
    selfuncMin = selfunc[0,0]
    selfuncMax = selfunc[-1,0]
    selfuncDx = selfunc[1,0] - selfunc[0,0]
    selfuncN = np.size(selfunc[:,0])
  
    selFunMin = max(selFunMin, selfuncMin)
    selFunMax = min(selFunMax, selfuncMax)
  
  
    def f(z): return selfunc[np.ceil((z-selfuncMin)/selfuncDx), 1]*z**2 
    def fTotal(z): return selfuncUnity[np.ceil((z-selfuncMin)/selfuncDx), 1]*z**2 
  
    zrange = np.linspace(selFunMin, selFunMax)
  
    nbar = scipy.integrate.quad(f, selFunMin, selFunMax)
    nbar = nbar[0]
  
    ntotal = scipy.integrate.quad(fTotal, 0.0, max(selfuncUnity[:,0]))
    ntotal = ntotal[0]
  
    nbar = ntotal / area / nbar


  return (volume, nbar)


# -----------------------------------------------------------------------------
# returns the nside resolution from the given maskfile
def getNside(maskFile):

  mask = healpy.read_map(maskFile)
  return healpy.get_nside(mask)


# -----------------------------------------------------------------------------
# helper function to convert RA,dec to phi,theta
def convertAngle(RA, Dec):

    phi   = np.pi/180.*RA
    theta = Dec*np.pi/180.
    theta = np.pi/2. - Dec*np.pi/180.

  return phi, theta

# -----------------------------------------------------------------------------
# computes the mask from a given galaxy datafile and writes it to a file
def figureOutMask(galFile, nside, outMaskFile):

  npix = healpy.nside2npix(nside)
  mask = np.zeros((npix))

  for line in open(galFile):
    line = line.split()
    RA  = np.float(line[3])
    Dec = np.float(line[4])
    z   = np.float(line[5])

    phi, theta = convertAngle(RA, Dec)

    pix = healpy.ang2pix(nside, theta, phi)
    mask[pix] = 1.
 
    healpy.write_map(outMaskFile, mask)

  return mask

# -----------------------------------------------------------------------------
# figures out which galaxies live on a mask edge, and also writes the edge
#   map to an auxillary file
def findEdgeGalaxies(galFile, maskFile, edgeGalFile, edgeMaskFile,
                     zmin, zmax, omegaM, useComoving=False):

  if useComoving:
    zmin = LIGHT_SPEED/100.*angularDiameter(zmin, Om=omegaM)
    zmax = LIGHT_SPEED/100.*angularDiameter(zmax, Om=omegaM)
  else:
    zmin *= LIGHT_SPEED/100.
    zmax *= LIGHT_SPEED/100.


  mask = healpy.read_map(maskFile)
  nside = healpy.get_nside(mask)
  npix = healpy.nside2npix(nside)
  edgeMask = np.zeros((npix))

  edgeFile = open(edgeGalFile, "w")

  for line in open(galFile):
    line = line.split()
    RA  = np.float(line[3])
    Dec = np.float(line[4])
    z   = np.float(line[5])

    if useComoving:
      z = LIGHT_SPEED/100.*angularDiameter(z, Om=omegaM)
    else:
      z *= LIGHT_SPEED/100.

    phi, theta = convertAngle(RA, Dec)

    isOnEdge = False

    # check the mask edges
    neighbors = healpy.get_all_neighbors(nside, theta, phi)
    isOnEdge = any(p == 0 for p in neighbors):
  
    # check the redshift boundaries

    if isOnEdge:
      edgeFile.write("1")
    else:
      edgeFile.write("0")

  edgeFile.close()
  healpy.write_map(edgeMaskFile, edgeMask)

  return
